# -*- coding: utf-8 -*-
"""
Unit library parse extension.

The methods here parse a string containing physical
quantities to numerical value and unit. The unit
manipulation and conversion is done using
Astropy unit library.
"""
import warnings

from astropy import units as u
from astropy.units import cds
from astropy.units import Quantity
from numpy import inf

__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["string_to_quantity", "ScalarQuantity"]

NUMERIC = "0123456789-+.eE*/j^ ()"

tr = u.def_unit(["tr", "turn", "revolution"], 1.0 * u.cycle)
u.add_enabled_units([cds.ppm, tr])

convert = {
    "Å": "Angstrom",
    "°C": "deg_C",
    "°F": "deg_F",
    "°": "deg",
    "µ": "u",
    "ℏ": "/h",
    "Ω": "Ohm",
}


def get_start_index_of_quantity_unit(string):
    """Return the index where the unit of the quantity starts."""
    string = string.strip() + " "

    for i, c in enumerate(string):
        if c not in NUMERIC:
            break

    j = 1
    for j in range(1, i + 1):
        if string[:i][-j] == "(":
            break
        if string[:i][-j] == ")":
            j = j - 1
            break

    index = i if i == j else i - j
    return index


def string_to_quantity(string, dtype=float):
    """
    Parse the string and return a ``Quantity`` object.

    The string must contain a physical quantity. The method parse the string
    for numerical value and unit.

    :returns: ``Quantity`` object
    """
    index = get_start_index_of_quantity_unit(string)

    # the string has numerical value and unit
    if index != -1:
        try:
            number = eval(string[:index])
        except ZeroDivisionError:
            number = inf
        except Exception as e:
            raise ValueError(e)
    # the string is unit only
    else:
        index = 0
        number = 1.0

    unit = string[index:].strip()
    if unit != "" and unit != "()":
        if unit[0] == "(" and unit[-1] == ")":
            unit = unit[1:-1]

    for key in convert:
        unit = unit.replace(key, convert[key])
        unit_multiplier = 1
    try:
        unit_qt = u.Unit(unit) * unit_multiplier
        analysis = dtype(number) * unit_qt
        return analysis
    except BaseException as e:
        raise BaseException(e)


def scalar_quantity_format(quantity, numerical_value=True):
    """Convert unit to value object."""
    string = quantity.unit.to_string("fits").strip()
    for key in convert.keys():
        string = string.replace(convert[key], key)
    lst = {"10**-6": "ppm", "10**-2": "%"}
    for key in lst.keys():
        string = string.replace(key, lst[key])

    if numerical_value:
        cat_string = [str(quantity.value)]
    else:
        cat_string = [""]
    subunits = string.split(" ")
    for item in subunits:
        power = False
        if item.find("-") == -1:
            for i, c in enumerate(item):
                if c.isnumeric():
                    if i == 0:
                        break
                    cat_string.append(item[:i] + "^" + item[i:] + " *")
                    power = True
                    break
            if not power:
                cat_string.append(item + " *")
        else:
            l, r = item.split("-")
            cat_string.append(l + "^-" + r + " *")
    string = " ".join(cat_string)[:-2]
    string = string.replace("* / *", "/")
    return string.strip()


# def unit_to_latex(unit):
#     """NotImplemented."""
#     string = unit.to_string("fits").strip()
#     convert_tex = {
#         "Angstrom": "\\AA",
#         "deg_C": "$^\\circ$C",
#         "deg_F": "$^\\circ$C",
#         "deg": "$^\\circ$",
#         "u": "$\\mu$",
#         "/h": "$\\hbar$",
#         "Ohm": "$\\Ohm$",
#         "10**-6": "ppm",
#     }
#     for key in convert_tex:
#         string = string.replace(key, convert_tex[key])
#     lst = {"10**-6": "ppm", "10**-2": "%"}
#     for key in lst.keys():
#         string = string.replace(key, lst[key])

#     cat_string = []
#     subunits = string.split(" ")
#     for item in subunits:
#         power = False
#         if item.find("-") == -1:
#             for i, c in enumerate(item):
#                 if c.isnumeric():
#                     if i == 0:
#                         break
#                     cat_string.append(item[:i] + "$^{" + item[i:] + "}$ ")
#                     power = True
#                     break
#             if not power:
#                 cat_string.append(item + " ")
#         else:
#             l, r = item.split("-")
#             cat_string.append(l + "$^{-" + r + "}$ ")
#     string = " ".join(cat_string)
#     return string


class ScalarQuantity:

    """
    A ScalarQuantity class compliant with the CSDM standards.

    @params: quantity: Return the quantity object from astropy.units
                       library.
    """

    __slots__ = "quantity"

    def __init__(self, quantity_string=None, unit=None):
        self.quantity = self.quantity_object(quantity_string, unit)

    @classmethod
    def quantity_object(cls, quantity_string, unit):
        if isinstance(quantity_string, ScalarQuantity):
            return quantity_string.quantity

        if isinstance(quantity_string, Quantity):
            return quantity_string

        lst = [None, ""]
        if quantity_string in lst and unit in lst:
            return 0.0 * u.Unit("")

        if quantity_string in lst and unit not in lst:
            return 0.0 * unit

        if isinstance(quantity_string, str):
            quantity = string_to_quantity(quantity_string)
            if unit is not None:
                return check_unit_consistency(quantity, unit)
            else:
                return quantity

    def __str__(self):
        return self.__format__()

    def __format__(self, format="quantity"):
        """Format the units according to csdmpy recommendation.

        :ivar: format: The value can either be 'quantity' or 'unit'

        .. doctest::

            >>> a = ScalarQuantity('5 kg m^2 /s')
            >>> print(a.quantity)
            5.0 kg m2 / s
            >>> print(str(ScalarQuantity(a.quantity)))
            5.0 kg * m^2 * s^-1
            >>> print(ScalarQuantity(a.quantity).__format__('unit'))
            kg * m^2 * s^-1
        """

        element = _default_units(self.quantity)
        if format == "unit":
            return scalar_quantity_format(element, numerical_value=False)
        if format == "quantity":
            return scalar_quantity_format(element)


def check_unit_consistency(element, unit):
    if isinstance(unit, str):
        unit = ScalarQuantity(unit).quantity.unit
    if element.unit.physical_type != unit.physical_type:
        options = [
            str(element.unit),
            str(element.unit.physical_type),
            str(unit),
            unit.physical_type,
        ]
        message = (
            "Validation Failed: The unit '{0}' ({1}) is inconsistent "
            "with the unit '{2}' ({3})."
        )
        raise Exception(message.format(*options))
    return element


def _default_units(element):
    if element.unit.physical_type == "frequency":
        element = element.to("Hz")
    return element


def check_quantity_name(element, unit):
    if element is None:
        element = unit.physical_type
        return element

    if unit.physical_type == "unknown":
        warnings.warn(
            f"The physical quantity name associated with the unit, {str(unit)}, is not"
            f" defined in astropy.units package. Continuing with '{element}' as the "
            f"physical quantity name."
        )
        return element

    if element.lower() != unit.physical_type:
        warnings.warn(
            f"The physical quantity name, '{element}', is not defined in the "
            f"astropy.units package. Continuing with '{element}' as the physical "
            f"quantity name for unit {str(unit)}."
        )

    return element.lower()


def frequency_ratio(offset):
    return [
        (u.Hz, u.cds.ppm, lambda x: (x / offset.to("Hz")) * 1e6, lambda x: x * offset)
    ]



# -*- coding: utf-8 -*-
"""Utility functions for the csdmpy module."""
import sys
from os import path
from urllib.parse import quote
from urllib.parse import urlparse

import requests


__author__ = "Deepansh J. Srivastava"
__email__ = "srivastava.89@osu.edu"
__all__ = ["parse_url", "download_file_from_url"]


def parse_url(url):
    res = urlparse(quote(url, safe="/?#@:"))
    return res


def download_file_from_url(url, verbose=False):
    res = parse_url(url)
    filename = path.split(res[2])[1]
    # name, extension = path.splitext(filename)
    # original_name = name
    # i = 0
    if path.isfile(filename):
        if verbose:
            sys.stdout.write(
                f"Found a local file with the filename, {0}. Skipping download."
            )
        return filename
        # filename = "{0}_{1}{2}".format(original_name, str(i), extension)
        # i += 1

    with open(filename, "wb") as f:
        response = requests.get(url, stream=True)
        total = response.headers.get("content-length")

        if total is None:
            f.write(response.content)
        else:
            downloaded = 0
            total = int(total)
            if verbose:
                sys.stdout.write(
                    "Downloading '{0}' from '{1}' to file '{2}'.\n".format(
                        res[2], res[1], filename
                    )
                )
            for data in response.iter_content(
                chunk_size=max(int(total / 1000), 1024 * 1024)
            ):
                downloaded += len(data)
                f.write(data)
                if verbose:
                    done = int(20 * downloaded / total)
                    sys.stdout.write("\r[{}{}]".format("█" * done, "." * (20 - done)))
                    sys.stdout.flush()
    if verbose:
        sys.stdout.write("\n")

    return filename


def _get_absolute_data_address(data_path, file):
    """
    Return the absolute path address of a local data file.

    :params: data_path:
    """
    _file_abs_path = path.abspath(file)
    _path, _file = path.split(_file_abs_path)
    _join = path.join(_path, data_path)
    _join = path.normpath(_join)
    return "file:" + _join


def get_absolute_url_path(url, file):
    res = parse_url(url)
    path = res.geturl()
    if res.scheme in ["file", ""]:
        if res.netloc == "":
            path = _get_absolute_data_address(res.path, file)
    return path


def get_relative_url_path(dataset_index, filename):
    index = str(dataset_index)
    absolute_path = get_absolute_url_path("", filename)

    name = path.splitext(path.split(filename)[1])[0] + "_" + index + ".dat"

    relative_url_path = path.join("file:.", name)

    absolute_path = path.abspath(
        urlparse(path.join(absolute_path, urlparse(relative_url_path).path)).path
    )
    return relative_url_path, absolute_path



# -*- coding: utf-8 -*-
"""Test for the csdm object
    1) T
"""
import numpy as np

import csdmpy as cp

data = np.random.rand(15, 5, 10)
a = cp.new()
dim1 = {"type": "linear", "count": 10, "increment": "1"}
dim2 = {"type": "linear", "count": 5, "increment": "1"}
dim3 = {"type": "linear", "count": 15, "increment": "1"}
dv = {"type": "internal", "components": [data.ravel()], "quantity_type": "scalar"}

a.add_dimension(dim1)
a.add_dimension(dim2)
a.add_dimension(dim3)
a.add_dependent_variable(dv)

data1 = np.random.rand(15 * 5 * 10).reshape(15, 5, 10) * 1e3 + 10
a1 = cp.new()
dv = {"type": "internal", "components": [data1.ravel()], "quantity_type": "scalar"}

a1.add_dimension(dim1)
a1.add_dimension(dim2)
a1.add_dimension(dim3)
a1.add_dependent_variable(dv)

# ---------------------------------------------------------------
# test for ufunc that apply to dimensionless dependent variables.
# __ufunc_list_dimensionless_unit__


def test_T():
    b = a.T
    out = a.dependent_variables[0].components[0]
    assert np.allclose(b.dependent_variables[0].components[0], out.T)
    b.save("test_abc.csdf")

    b_ = cp.load("test_abc.csdf")
    assert b == b_



# -*- coding: utf-8 -*-
import pytest

from csdmpy.utils import check_and_assign_bool
from csdmpy.utils import check_encoding
from csdmpy.utils import NumericType
from csdmpy.utils import QuantityType


def test_encoding():
    assert check_encoding("base64") == "base64"
    assert check_encoding("raw") == "raw"
    assert check_encoding("none") == "none"

    error = "is an invalid `encoding` enumeration literal. The allowed values are"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        check_encoding("text")


def test_quantity_type():
    assert QuantityType("pixel_4").value == "pixel_4"
    assert QuantityType("pixel_4").p == 4
    assert QuantityType("scalar").value == "scalar"
    assert QuantityType("vector_15").value == "vector_15"
    assert QuantityType("vector_15").p == 15
    assert QuantityType("matrix_13_3").value == "matrix_13_3"
    assert QuantityType("matrix_13_3").p == 39
    assert QuantityType("symmetric_matrix_10").value == "symmetric_matrix_10"
    assert QuantityType("symmetric_matrix_10").p == 55

    error = "is an invalid `quantity_type` enumeration literal. The allowed values are"
    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        QuantityType("RGB")


def test_numeric_type():
    assert NumericType("uint8").value == "uint8"
    assert NumericType("uint8").dtype == "<u1"
    assert NumericType("uint16").value == "uint16"
    assert NumericType("uint16").dtype == "<u2"
    assert NumericType("uint32").value == "uint32"
    assert NumericType("uint32").dtype == "<u4"
    assert NumericType("uint64").value == "uint64"
    assert NumericType("uint64").dtype == "<u8"

    assert NumericType("int8").value == "int8"
    assert NumericType("int8").dtype == "<i1"
    assert NumericType("int16").value == "int16"
    assert NumericType("int16").dtype == "<i2"
    assert NumericType("int32").value == "int32"
    assert NumericType("int32").dtype == "<i4"
    assert NumericType("int64").value == "int64"
    assert NumericType("int64").dtype == "<i8"

    # assert NumericType("float16").value == "float16"
    # assert NumericType("float16").dtype == "<f2"
    assert NumericType("float32").value == "float32"
    assert NumericType("float32").dtype == "<f4"
    assert NumericType("float64").value == "float64"
    assert NumericType("float64").dtype == "<f8"

    assert NumericType("complex64").value == "complex64"
    assert NumericType("complex64").dtype == "<c8"
    assert NumericType("complex128").value == "complex128"
    assert NumericType("complex128").dtype == "<c16"

    error = "is an invalid `numeric_type` enumeration literal. The allowed values are"

    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        NumericType("float16")

    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        NumericType("complex256")

    with pytest.raises(ValueError, match=".*{0}.*".format(error)):
        NumericType("float128")


def test_boolean():
    assert check_and_assign_bool(None) is False
    assert check_and_assign_bool(False) is False
    assert check_and_assign_bool(True) is True

    error = "Expecting an instance of type"
    with pytest.raises(TypeError, match=".*{0}.*".format(error)):
        check_and_assign_bool("True")



# -*- coding: utf-8 -*-
from csdmpy.units import ScalarQuantity


def test_unit():
    a = ScalarQuantity("1 tr").quantity
    assert str(a) == "1.0 tr"
    assert str(a.to("cycle")) == "1.0 cycle"
    # assert ScalarQuantity(a).format() == '1.0 tr'

    a = ScalarQuantity("10 cm^-1/s")
    assert str(a) == "10.0 cm^-1 * s^-1"
    assert ScalarQuantity(a).quantity == a.quantity

    a = ScalarQuantity("1 deg").quantity
    assert str(a) == "1.0 deg"
    assert str(ScalarQuantity(a)) == "1.0 °"
    assert str(ScalarQuantity("1 deg")) == "1.0 °"

    a = ScalarQuantity("(54.3/2) ppm").quantity
    assert str(a) == "27.15 ppm"
    assert str(ScalarQuantity(a)) == "27.15 ppm"

    a = ScalarQuantity("(54.3/2) (µHz/Hz)").quantity
    assert str(a) == "27.15 uHz / Hz"
    assert str(ScalarQuantity(a)) == "27.15 Hz^-1 * µHz"
    assert str(a.to("ppm")) == "27.15 ppm"

    a = ScalarQuantity("5 kg * m / s").quantity
    b = a * ScalarQuantity("1 m / s").quantity
    assert str(a) == "5.0 kg m / s"
    assert str(ScalarQuantity(a)) == "5.0 kg * m * s^-1"
    assert str(ScalarQuantity(b)) == "5.0 kg * m^2 * s^-2"
    assert ScalarQuantity(b).__format__("unit") == "kg * m^2 * s^-2"
    assert str(b.to("J")) == "5.0 J"

    a = ScalarQuantity("5e-7 s").quantity
    assert str(a) == "5e-07 s"
    assert str(ScalarQuantity(a)) == "5e-07 s"
    assert str(ScalarQuantity(a.to("us"))) == "0.5 µs"
    assert ScalarQuantity(a.to("us")).__format__("unit") == "µs"

    a = ScalarQuantity("5e-7 * 2 s").quantity
    assert str(a) == "1e-06 s"
    assert str(ScalarQuantity(a)) == "1e-06 s"
    assert str(ScalarQuantity(a.to("us"))) == "1.0 µs"



# -*- coding: utf-8 -*-
from os.path import split

import numpy as np
import pytest

import csdmpy as cp


def test_load_files():
    assert split(cp.tests.test01)[1] == "yis83dhuj1q7v2srw2b1hjnb66d1rw1h.csdf"
    assert split(cp.tests.test02)[1] == "0jsvs9bmme8vf6yn76bctkhfxm4vkcc8.csdf"


def test00():
    error = "Missing the value for the required `filename` attribute."
    with pytest.raises(Exception, match=".*{0}.*".format(error)):
        cp.load()


def test01():
    dataset = cp.load(cp.tests.test01)

    assert dataset.dependent_variables[0].type == "internal"

    # encoding is always set to 'base64' after import
    assert dataset.dependent_variables[0].encoding == "base64"

    assert dataset.dependent_variables[0].numeric_type == "float32"

    assert dataset.dependent_variables[0].components.dtype == np.float32

    assert dataset.description == "A simulated sine curve."

    assert len(dataset.dependent_variables) == 1

    assert len(dataset.dimensions) == 1

    assert dataset.dimensions[0].type == "linear"

    assert str(dataset.dimensions[0].increment) == "0.1 s"

    assert str(dataset.dimensions[0].origin_offset) == "0.0 s"

    assert dataset.dimensions[0].count == 10

    assert dataset.dimensions[0].quantity_name == "time"

    assert np.all(dataset.dimensions[0].coordinates.value == np.arange(10) * 0.1)


def test02():
    dataset = cp.load(cp.tests.test02)

    assert dataset.dependent_variables[0].type == "internal"

    # encoding is always set to 'base64' after import
    assert dataset.dependent_variables[0].encoding == "base64"

    assert dataset.dependent_variables[0].numeric_type == "float64"

    assert dataset.dependent_variables[0].components.dtype == np.float64

    assert dataset.description == "Base64 encoding test"

    assert len(dataset.dependent_variables) == 4

    assert len(dataset.dimensions) == 1

    assert dataset.dimensions[0].type == "monotonic"

    assert str(dataset.dimensions[0].origin_offset) == "0.0 cm"

    assert dataset.dimensions[0].count == 10

    assert dataset.dimensions[0].quantity_name == "length"


# def test03():
#     dataset1 = cp.load(cp.tests.test03, sort_fft_order=False)
#     dataset2 = cp.load(cp.tests.test03, sort_fft_order=True)

#     dat1 = dataset1.dependent_variables[0].components[0]
#     dat2 = dataset2.dependent_variables[0].components[0]
#     assert np.all(np.fft.fftshift(dat1) == dat2)


def test04():
    data1 = cp.load(cp.tests.test04)
    data2 = data1.copy()
    assert data1.data_structure == data2.data_structure
